;;; tramp-file-name.el --- Test file name mangling operations

;; Copyright (C) 2001 by Daniel Pittman <daniel@localhost>

;;; Code:

(push "." load-path)
(push "../lisp" load-path)
(load "tramp2.el")


;; Make sure that './' constructs come through ok...



;;; tramp-file-name.el ends here
